/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mtk_cubic_image.h"
#include  "../common/math/vector3.h"
#include  "../common/math/quaternion.h"
#include  "../common/graphics/gl_image.h"
#include  "../common/astro/astrometry.h"
#include  "mtk_draw_struct.h"
#include  "mtk_camera_setting.h"
#include  "mtk_pbuffer.h"
#include  "mtk_camera.h"
#include  "mtk_make_movie.h"

#include  <direct.h>



//--  Externs
mtk_draw_struct		calc_draw_struct(int w, int h);
void				draw_scene(mtk_draw_struct ds);


extern char tyTarget;
extern mtk_camera	g_camera;
extern int gl_w, gl_h;


struct axes {
	vector3f  eF, eU;
};






bool
make_cubic_image(const char * base_filename, int size)
{
	const char szFace[6][4] = { "LFT", "RGT", "TOP", "BTM", "CTR", "BCK" };

	const  axes Axes[6] = {
		{ vector3f(0, 1,0), vector3f(0,0,1)  },		// L
		{ vector3f(0,-1,0), vector3f(0,0,1)  },		// R
		{ vector3f(0,0, 1), vector3f(-1,0,0) },		// U
		{ vector3f(0,0,-1), vector3f( 1,0,0) },		// D
		{ vector3f( 1,0,0), vector3f(0,0,1)  },		// F
		{ vector3f(-1,0,0), vector3f(0,0,1)  }		// B
	};
	


	int w, h;
	w = h = size;
	if ( !create_pbuffer(size, size) )  return false;

	matrix3d  M = rotation_matrix3(M_PI, vector3d(1,0,0)) * M_R2E;
	g_camera.set_ori( matrix3_to_quaternion(M) );

	image  img(w,h,3);
	glViewport(0,0,w,h);
	for (int i=0; i<6; i++) {
		const vector3f& eF = Axes[i].eF;
		const vector3f& eU = Axes[i].eU;
		quaternionf  ori = axes_to_quaternion(eF, eU);

		set_local_camera(ori, vector3f(0,0,0));

		//-- Clear
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );


		draw_scene( calc_draw_struct(w, h) );
		gl_read_pixels(img, 0,0, w, h);

		char fn[_MAX_PATH];
		sprintf(fn, "%s_%s.png", base_filename, szFace[i]);
		img.savePNG(fn);
	}
	destroy_pbuffer();

	return  true;
}




bool
make_cubic_frame(const char* mv_basefilename, float t, int frame, int size, HDC hDC)
{
	const char szFace[6][4] = { "LFT", "RGT", "TOP", "BTM", "CTR", "BCK" };
	const char szEye[] = { 'L', 'R' };

	const  axes Axes[6] = {
	{ vector3f(-1, 0, 0), vector3f( 0, 1, 0) },	// L
	{ vector3f( 1, 0, 0), vector3f( 0, 1, 0) },	// R
	{ vector3f( 0, 1, 0), vector3f( 0, 0, 1) },	// U
	{ vector3f( 0,-1, 0), vector3f( 0, 0,-1) },	// D
	{ vector3f( 0, 0,-1), vector3f( 0, 1, 0) },	// F
	{ vector3f( 0, 0, 1), vector3f( 0, 1, 0) }	// B
	};


	int w, h;
	w = gl_w;
	h = gl_h;

	if (_chdir(".\\cubic_movie")) {
		_mkdir(".\\cubic_movie");
		_chdir(".\\cubic_movie");
	}

	quaterniond base_ori(0.985, -0.174, 0.0, 0.0);
	base_ori.normalize();
	set_base_camera( base_ori );

	const vector3f rot_axis(0, 0.940, 0.342);
	const float rot_angle = 0.401 / 180.0 * M_PI;
	const float eye_dist = 7;

	image  img(w,h,3);
	glViewport(0,0,w,h);
	for (int i=0; i<6; i++) {
		const vector3f& eF = Axes[i].eF;
		const vector3f& eU = Axes[i].eU;
		quaternionf  ori = axes_to_quaternion(eF, eU);

		for (int j=0; j<2; j++) {
			char dir[256];
			char filename[256];

			sprintf(dir, "%s_%c", szFace[i], szEye[j]);
			sprintf(filename, "%s_%s_%c_%05d.tga", mv_basefilename, szFace[i], szEye[j], frame);

			if (_chdir(dir)==-1) {
				_mkdir(dir);
				_chdir(dir);
			}

			float eye_x = (j==0)  ?  -0.5*eye_dist  :  0.5*eye_dist;
			float ang   = (j==0)  ?  -rot_angle  :  rot_angle;

			set_local_camera( make_rotation_quaternion(ang, rot_axis) * ori, vector3f(eye_x,0,0));

			glPushMatrix();
				draw_intr_frame(t);
			glPopMatrix();

			gl_read_pixels(img, 0,0, w, h);
			img.saveTGA(filename);

			SwapBuffers(hDC);

			_chdir("..");
		}
	}
	_chdir("..");
	set_base_camera( quaterniond(1,0,0,0) );

	return  true;
}
