/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mtk_object.h"
#include  "mtk_object_position.h"
#include  "mtk_planets.h"
#include  "mtk_planet_position.h"
#include  "mtk_planet_orientation.h"
#include  "mtk_planet_phys_params.h"
#include  "mtk_minor_planets.h"
#include  "mtk_oort_cloud.h"
#include  "mtk_fixed_object.h"
#include  "mtk_stars.h"
#include  "mtk_galaxies.h"
#include  "mtk_spacecraft.h"
#include  "mtk_label.h"
#include  "mtk_scale_line.h"
#include  "mtk_switches.h"
#include  <vector>
#include  <algorithm>


void
mtk_object::init(int type, int idx, float dz)
{
	m_type = type;
	m_idx  = idx;
	m_dz   = dz;
}


vector3d
mtk_object::get_pos() const
{
	return  get_object_positon( *this );
}




bool
mtk_object::is_visible() const
{
	switch (m_type) {  //****
		case OBJ_TYPE_PLANET: 
			if (m_idx == 8)			return  test_disp_switch( DISPSW_TNO,		DISPM_DISP);	//--  Pluto
			else					return  test_disp_switch( DISPSW_PLANET,	DISPM_DISP);
		case OBJ_TYPE_SATELLITE:	return  test_disp_switch( DISPSW_SATELLITE,	DISPM_DISP);
		case OBJ_TYPE_SMALL_OBJECT:
			if (get_small_object_type(m_idx) == SMALLOBJ_TYPE_MINOR_PLANETS) {
				return  test_disp_switch( DISPSW_MINOR_PLANET,	DISPM_DISP);
			} else {
				return  test_disp_switch( DISPSW_TNO,	DISPM_DISP);
			}

		case OBJ_TYPE_OORT_CLOUD:	return  bOortCloud;
		case OBJ_TYPE_SCALE_CIRCLE:	return  false;
		
		default:				return  true;
	}
}



double
mtk_object::get_bounding_R(const mtk_draw_struct& ds) const
{
	switch (m_type) {
		case OBJ_TYPE_SUN:			return  get_zoom_rate().zmhSun * get_sun_glare_radius();
		case OBJ_TYPE_PLANET:		return  2 * get_planet_Re(m_idx) * _1km_pc;		//***
		case OBJ_TYPE_SATELLITE:
			{
				double fac = (m_idx==SAT_MOON)  ?  get_zoom_rate().zmhPlanet  :  1.0;
				double Rx, Ry, Rz;
				get_satellite_size(m_idx, Rx, Ry, Rz);
				return  _1km_pc * Rx * fac;
			}

		case OBJ_TYPE_SMALL_OBJECT:
			return  1000 * _1km_pc;	//***

		case OBJ_TYPE_OORT_CLOUD:	return  get_Oort_cloud_radius();

		case OBJ_TYPE_LABEL:
			{
				double L = norm( get_object_positon( *this ) - ds.camera_base.get_pos() );

				// TCY in pc
				float sz = get_label_letter_size(m_idx) * L/screen_z * screen_w / 180.0;
				int len = get_label_letter_count( m_idx, ds.lang == LANG_ENG ); 

				return sz*len;	//***
			}

		//******//
		default:
			return  100 * _1km_pc;
	}
}




void
mtk_object::draw(const mtk_draw_struct& ds)  const
{
	// IuWFNg̃J̑Έʒu
	vector3d  rpos = get_pos() - ds.camera.get_pos();

	// XP[̌vZ
	double  scl;
	if (tyMode == MODE_PLANETARIUM) {
		const double Clst_R = 1.0023;
		scl = Clst_R / rpos.norm();
	} else {
		scl = 1.0;
		if (!ds.is_near_region()) {
			double l = rpos.norm() * ds.scale_factor;
			if (l > z_far2 * 0.8) {
				scl = z_far2 * 0.8 / l;
			}
		}
	}
	rpos *= scl;


	switch (m_type) {
		case OBJ_TYPE_SUN:				draw_sun(ds, 0, rpos, scl);					break;
		case OBJ_TYPE_PLANET:			draw_a_planet_3D(ds, m_idx, rpos, scl);		break;
		case OBJ_TYPE_SATELLITE:		draw_a_satellite(ds, m_idx, rpos, scl);		break;
		case OBJ_TYPE_SMALL_OBJECT:		draw_a_small_object(ds, m_idx, rpos, scl);	break;
		case OBJ_TYPE_OORT_CLOUD:		draw_Oort_cloud(ds, 0, rpos, scl);			break;
		case OBJ_TYPE_LABEL:			draw_a_label(ds, m_idx, rpos, scl, m_dz);	break;
		default:	break;
	}
}




//==========
//  Object
//==========
class z_list
{
public:
	int		idx;	// idx
	float	z;		// z-l

	bool operator<(const z_list& zl) const  { return z < zl.z; }
};


static std::vector<mtk_object>	vObject;
static std::vector<z_list>		vZ;



void
clear_objects()
{
	vObject.clear();
	vZ.clear();
}

void
register_object(const mtk_object& obj)
{
	vObject.push_back(obj);
}



void
init_objects()
{
	register_solar_system_objects();
}






//-----------
//  culling  
//-----------
void
make_culling_info(const frustum& frs, vector3d nv[4])
{
	float x1, x2, y1, y2, z1, z2;
	x1 =  frs.get_x1();
	x2 =  frs.get_x2();
	y1 =  frs.get_y1();
	y2 =  frs.get_y2();
	z1 = -frs.get_z1();

	//--
	vector3d  v11, v12, v21, v22;
	v11.set(x1, y1, z1);
	v12.set(x2, y1, z1);
	v21.set(x1, y2, z1);
	v22.set(x2, y2, z1);

	
	// eʂ̊O@xNg߂
	nv[0] = normalize( cross_product(v21, v11) );	//  Left
	nv[1] = normalize( cross_product(v12, v22) );	//  Right
	nv[2] = normalize( cross_product(v22, v21) );	//  Top
	nv[3] = normalize( cross_product(v11, v12) );	//  Bottom
}



// \̈ɂꍇ true
inline  bool
cull_object(const mtk_draw_struct& ds, const vector3d& pos, const double& R, vector3d nv[], bool bCullFarSide)
{
	//--  _Wnł̈ʒu
	vector3d  rv = ds.calc_pos_on_viewframe(pos);

	if (rv.z - R > -ds.frs.get_z1())  return  false;
	if (bCullFarSide) {
		if (rv.z + R < -ds.frs.get_z2())  return  false;
	}

	for (int i=0; i<4; i++) {
		double d = dot_product(nv[i], rv);
		if (d > R)  return  false;
	}

	return  true;
}



//---  make z-sorted list  ---
int
make_z_list(const mtk_draw_struct& ds)
{
	const int N = vObject.size();
	if (vZ.size() < N) {
		vZ.resize(N);
	}

	vector3d  nv[4];
	make_culling_info(ds.frs, nv);

	const bool bCullFarSide = ds.is_near_region();	//***
	const double sf = ds.scale_factor;

	int num = 0;
	for (int i=0; i<N; i++) {
		const mtk_object& obj = vObject[i];

		if ( !obj.is_visible() )  continue;

		// Eɂ̂̂ݓo^
		vector3d  pos = obj.get_pos();
		double Rb = obj.get_bounding_R(ds);

		if ( !cull_object(ds, pos, sf*Rb, nv, bCullFarSide) )  continue;

		vector3d  rv = ds.calc_pos_on_viewframe( obj.get_pos() );

		vZ[num].idx = i;
		vZ[num].z   = -rv.z + obj.get_dz();
		num++;
	}

	sort(vZ.begin(), vZ.begin() + num);

	return  num;
}






void
draw_objects(const mtk_draw_struct& ds)
{
	int num;
	num = make_z_list(ds);

	//---  `悷  ---
	for (int i=num-1; i>=0; i--) {
		vObject[ vZ[i].idx ].draw(ds);
	}
}





//
//
//
bool
is_object_label_visible(const mtk_object& obj)
{
	const int obj_type = obj.get_type();
	const int idx = obj.get_index();
	char dispsw_idx;



	switch (obj_type) {
		//--  Sun
		case OBJ_TYPE_SUN:
			return  test_disp_switch(DISPSW_PLANET, DISPM_NAME);
		//--  Planets
		case OBJ_TYPE_PLANET:
			if (idx == 8)	dispsw_idx = DISPSW_TNO;
			else			dispsw_idx = DISPSW_PLANET;
			break;

		case OBJ_TYPE_SATELLITE:
			dispsw_idx = DISPSW_SATELLITE;
			break;
		
		//--  Small Objects
		case OBJ_TYPE_SMALL_OBJECT:
			{
				int type = get_small_object_type( idx );
				if (type == SMALLOBJ_TYPE_TNOS)		dispsw_idx = DISPSW_TNO;
				else								dispsw_idx = DISPSW_MINOR_PLANET;
			}
			break;

		//--  Oort cloud
		case OBJ_TYPE_OORT_CLOUD:
			return  bOortCloud && test_disp_switch(DISPSW_PLANET, DISPM_NAME);

		//--  Fixed label
		case OBJ_TYPE_FIXED_OBJECT:
			return  is_fixed_object_label_visible( idx );

		case OBJ_TYPE_SCALE_CIRCLE:
			if (swScaleLine & SCALE_SQUARE && idx == 20) {
				return  true;
			}
			return  swScaleLine & SCALE_CIRCLE;

		default:
			return  true;
	}


	if ( !test_disp_switch( dispsw_idx, DISPM_DISP) || !test_disp_switch( dispsw_idx, DISPM_NAME ) )  return  false;


	if (obj_type == OBJ_TYPE_SATELLITE) {
		if (test_disp_switch(dispsw_idx, DISPM_SELECT) && !is_satellite_selected(idx)) {
			return  false;
		}
	}

	if (obj_type == OBJ_TYPE_SMALL_OBJECT) {
		if (test_disp_switch(dispsw_idx, DISPM_SELECT) && !is_small_object_selected(idx)) {
			return  false;
		}
	}
	return  true;
}




bool
is_object_orbit_visible(const mtk_object& obj)
{
	const int obj_type = obj.get_type();
	const int idx = obj.get_index();
	char dispsw_idx;


	switch (obj_type) {
		//--  Planets
		case OBJ_TYPE_PLANET:
			if (idx == 8)	dispsw_idx = DISPSW_TNO;
			else			dispsw_idx = DISPSW_PLANET;
			break;

		case OBJ_TYPE_SATELLITE:
			dispsw_idx = DISPSW_SATELLITE;
			break;
		
		//--  Small Objects
		case OBJ_TYPE_SMALL_OBJECT:
			{
				int type = get_small_object_type( idx );
				if (type == SMALLOBJ_TYPE_TNOS)		dispsw_idx = DISPSW_TNO;
				else								dispsw_idx = DISPSW_MINOR_PLANET;
			}
			break;


		case OBJ_TYPE_SCALE_CIRCLE:
			if (swScaleLine & SCALE_SQUARE && idx == 20) {
				return  true;
			}
			return  swScaleLine & SCALE_CIRCLE;

		default:
			return  true;
	}


	if ( !test_disp_switch( dispsw_idx, DISPM_DISP) || !test_disp_switch( dispsw_idx, DISPM_ORBIT ) )  return  false;


	if (obj_type == OBJ_TYPE_SATELLITE) {
		if (test_disp_switch(dispsw_idx, DISPM_SELECT) && !is_satellite_selected(idx)) {
			return  false;
		}
	}

	if (obj_type == OBJ_TYPE_SMALL_OBJECT) {
		if (test_disp_switch(dispsw_idx, DISPM_SELECT) && !is_small_object_selected(idx)) {
			return  false;
		}
	}


	return  true;
}





//---------
//  Alpha
//---------
float
get_label_alpha(const mtk_draw_struct& ds, const mtk_object& obj)
{
	int idx = obj.get_index();

	switch (obj.get_type()) {
		case OBJ_TYPE_SUN:			return  get_sun_label_alpha( ds );
		case OBJ_TYPE_PLANET:		return  get_planet_label_alpha( ds, idx );
		case OBJ_TYPE_SATELLITE:	return  get_satellite_label_alpha( ds, idx );
		case OBJ_TYPE_SMALL_OBJECT:	return  get_small_object_label_alpha( ds, idx );
		case OBJ_TYPE_OORT_CLOUD:	return  get_Oort_cloud_alpha( ds );
		case OBJ_TYPE_FIXED_OBJECT:	return	get_fixed_object_label_alpha( ds, idx );
		case OBJ_TYPE_SCALE_CIRCLE:	return  get_scale_circle_alpha( ds, idx );
		default:				break;
	}

	return  1.0f;
}

float
get_orbit_alpha(const mtk_draw_struct& ds, const mtk_object& obj)
{
	int idx = obj.get_index();

	switch (obj.get_type()) {
		case OBJ_TYPE_PLANET:		return  get_planet_orbit_alpha( ds, idx );
		case OBJ_TYPE_SATELLITE:	return  get_satellite_orbit_alpha( ds, idx );
		case OBJ_TYPE_SMALL_OBJECT:	return  get_small_object_orbit_alpha( ds, idx );
		case OBJ_TYPE_SCALE_CIRCLE:	return  get_scale_circle_alpha( ds, idx );
		default:				break;
	}

	return  1.0f;
}
