/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mitaka.h"
#include  "mtk_spacecraft.h"
#include  "mtk_spacecraft_model.h"
#include  "mtk_directory.h"
#include  "../common/graphics/3ds_loader.h"
#include  "mtk_gl_common.h"
#include  "mtk_font.h"
#include  "../common/astro/astro_const.h"
#include  "mtk_planet_position.h"
#include  "mtk_switches.h"


//***
//--  Externs
float	calc_alpha2(const mtk_draw_struct& ds, const double& r, const float rs1, const float rs2);
void	set_light_from_sun(const vector3d& pos);


//-----------
//  Objects
//-----------
static spacecraft  vSpacecraft[5];
static vector3d vSpacecraftPosition[5];
//---  3D Models  ---
static glModel3D  mdlPioneer, mdlVoyager, mdlCassini;


void
init_spacecraft()
{
	directory  dir = get_directory(DIR_SPACECRAFT);

	//-- Load models
	load_3DS_file( dir.get_path("Pioneer.3ds"), mdlPioneer);
	load_3DS_file( dir.get_path("Voyager.3ds"), mdlVoyager);
	load_3DS_file( dir.get_path("Cassini.3ds"), mdlCassini);

	mdlPioneer.fit_size(1.0);
	mdlVoyager.fit_size(1.0);
	mdlCassini.fit_size(1.0);
	mdlCassini.set_r0( vector3f(0,0,0.5) );

	//-- init spacecraft
	vSpacecraft[0].set( "Pioneer 10",	"pCIjAPO",	&mdlPioneer);
	vSpacecraft[1].set( "Pioneer 11",	"pCIjAPP",	&mdlPioneer);
	vSpacecraft[2].set( "Voyager I",	"{CW[P",		&mdlVoyager);
	vSpacecraft[3].set( "Voyager II",	"{CW[Q",		&mdlVoyager);
	vSpacecraft[4].set( "Cassini",		"JbV[j",			&mdlCassini);

	//-- load path
	vSpacecraft[0].load_path( dir.get_path("Pioneer10.pth") );
	vSpacecraft[1].load_path( dir.get_path("Pioneer11.pth") );
	vSpacecraft[2].load_path( dir.get_path("Voyager_I.pth") );
	vSpacecraft[3].load_path( dir.get_path("Voyager_II.pth") );
	vSpacecraft[4].load_path( dir.get_path("Cassini.pth") );
}


//--------
//  Name
//--------
bilingual
get_spacecraft_name(int idx)
{
	return  vSpacecraft[idx].get_name();
}

//------------
//  Position
//------------
void
update_spacecraft_position(const double& t)
{
	for (int i=0; i<5; i++) {
		vSpacecraftPosition[i] = vSpacecraft[i].get_pos(t);
	}
}

vector3d
get_spacecraft_position(int idx)
{
	return  vSpacecraftPosition[idx];
}


//-------------
//  Draw Path
//-------------
bool
get_spacecraft_path_mode(int idx)
{
	return  vSpacecraft[idx].get_draw_path_mode();
}

void
set_spacecraft_path_mode(int idx, bool bFlg)
{
	vSpacecraft[idx].set_draw_path_mode(bFlg);
}





//---------------------
//  T@̃f̕\
//---------------------
void
draw_a_spacecraft(const mtk_draw_struct& ds, const spacecraft& spccrf)
{
	const vector3d  pos = spccrf.get_pos(ds.JD);		// T@̈ʒu

	// ̐ݒ
	set_light_from_sun(pos);

	// T@n̕Ɍs
	const vector3d  posE = get_planet_position(PLNT_EARTH);	// n̈ʒu
	vector3d nz = -normalize( posE - pos );
	vector3d nu = vector3d(0,0,1);
	vector3d nx = normalize( cross_product(nu, nz) );
	vector3d ny = normalize( cross_product(nz, nx) );

	double MR[16];
	MR[0] = nx.x;
	MR[1] = nx.y;
	MR[2] = nx.z;
	MR[3] = 0.0;

	MR[4] = ny.x;
	MR[5] = ny.y;
	MR[6] = ny.z;
	MR[7] = 0.0;

	MR[8]  = nz.x;
	MR[9]  = nz.y;
	MR[10] = nz.z;
	MR[11] = 0.0;

	MR[12] = 0.0;
	MR[13] = 0.0;
	MR[14] = 0.0;
	MR[15] = 1.0;

	vector3d rr = pos - ds.camera.get_pos();

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glPushMatrix();
		glTranslated(rr.x,rr.y,rr.z);
		glMultMatrixd(MR);
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_LIGHTING);
		glEnable(GL_NORMALIZE);
		glEnable(GL_LIGHTING);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);

		const float R = 100.0 * _1km_pc;
		glScalef(R, R, R);
		float M[16];
		glGetFloatv( GL_MODELVIEW_MATRIX, M);
		float L2 = M[12]*M[12] + M[13]*M[13] + M[14]*M[14];
		float R2 = M[0]*M[0] + M[1]*M[1] + M[2]*M[2];
		const float th1_2 = R2 / L2;
		const float thr1    = 0.05 * 3.14/180.0;
		const float thr2    = 1.0 * 3.14/180.0;
		const float thr1sqr = thr1*thr1;
		const float thr2sqr = thr2*thr2;

		// f̕\
		if (th1_2 > thr1sqr) {
			spccrf.draw_model();
		}

		// x̕\
		if ( test_disp_switch(DISPSW_PLANET, DISPM_NAME) && ds.JD>=spccrf.get_t1()) {
			float af;
			if (!spccrf.get_draw_path_mode()) {
				// O\ĂȂ́A߂Â̂ݖO\
				af = (th1_2 - thr1sqr) / (thr2sqr - thr1sqr);
			} else {
				// O\́AfƓ
				af = calc_alpha2( ds, norm(pos), 0.015, 0.03 );
			}
			if (af > 1.0f)  af = 1.0f;

			if (af > 0) {
				float sz = 2.5f * ds.ltf;
				glDisable(GL_TEXTURE_2D);
				glDisable(GL_LIGHTING);
				glEnable(GL_BLEND);
				glColor4f(0.65,0.65,0.70,af);
				prepare_3Dtext_1(1.4*R, 45, sz, ds.scale_factor, false);
				spccrf.draw_label(get_font(), (get_language()==LANG_ENG));
			}
		}
	glPopMatrix();
	glPopAttrib();
}

void
draw_spacecraft(const mtk_draw_struct& ds)
{
	for (int i=0; i<5; i++) {
		const spacecraft& spccrf = vSpacecraft[i];
		if (ds.JD < spccrf.get_t1())  continue;
		draw_a_spacecraft(ds, spccrf);
	}
}

void
draw_spacecraft_path(const mtk_draw_struct& ds)
{
	const vector3d camera_pos = ds.camera.get_pos();
	for (int i=0; i<5; i++) {
		const spacecraft& spccrf = vSpacecraft[i];
		float af = calc_alpha2( ds, norm( spccrf.get_pos(ds.JD) ), 0.015, 0.03 );
		if (af <= 0.0)  continue;
		RGBf  col;
		if (ds.JD <= spccrf.get_t2())	{ col.set(1,1,0); }
		else						{ col.set(0.7,0.5,0); }
		spccrf.draw_path(ds, ds.JD, camera_pos, col, af);
	}
}
